#!/usr/local/bin/perl -Tw
my $VERSION = "0.5a";

=head1 NAME

newsproxy.pl - a perl-based Usenet "proxy" server 

=head1 SYNOPSIS

newsproxy.pl [-hvd] [-p port]

=head1 DESCRIPTION

A news server written in perl, which gets its information from other news
servers.  Its primary purpose is to allow a standard newsreader to read
news from more than one site.  It also offers a way of reading the same
group in more than one different way.  Of course, its main purpose is to
be a front-end for Net::NNTP::Server and Client.

=cut

### User Variables ##########################################################
use vars qw( $PORT $HOME $DEBUG @MACHINES $MAXKIDS @CLIENTS );
# Which machines should we connect to?  For now, just put in
# Net::NNTP::Client references (we can take care of the authentication
# later if you'd like).
@CLIENTS = (
	new Net::NNTP::Client('news.cso.uiuc.edu',
		'server' => 'news-proxy.cso.uiuc.edu' ),
	new Net::NNTP::Client('news.stanford.edu'),
	   );

BEGIN {		# So that $HOME can be evaluated first, before 'use lib'
  $HOME = "/home/tskirvin/newsproxy";  # Where the files are located
  $PORT       = 9119;		  # 119 is NNTP, but it's reserved
  $DEBUG      = 0;		  # Print debugging information?
  $MAXKIDS    = 5; 		  # Maximum number of children running
  # Which machines can we cannect from?  List them off; regexps are fine.
  @MACHINES   = qw( localhost );
}

#############################################################################

=item USAGE

newsproxy.pl [-hvd] [-p port]

=over 4

=item -h

Prints a short help message and exits.

=item -v

Print the version number and exit.

=item -d

Print debugging information.

=item -p port

Bind to 'port' instead of the default ($PORT).  Intended to allow you to
bind to port 119 (the standard NNTP port); note that you will need root
priveleges to do this, since 119 is a priveleged port.

=back

=cut

use Net::NNTP;
use strict;
use Sys::Hostname;
use Getopt::Std;
use POSIX qw(:sys_wait_h :signal_h :errno_h);
use Errno;

use lib $HOME;
use Net::NNTP::Client;
use Net::NNTP::Server;
use News::NNTPAuth;

use vars qw( $CHILDREN $SERVER $opt_p $opt_d $opt_v $opt_h );
$CHILDREN = 0;
$SIG{CHLD} = \&reaper; 		   # Set what to do when the children die.

###############################################################################
##### main() ##################################################################
###############################################################################

# Trim the path name off of the program
$0 =~ s%.*/%%;

# Get the path information
getopts('p:dhv');
Usage()   if $opt_h;
Version() if $opt_v;
if ($opt_d) {
  $DEBUG ||= $opt_d;
  $Net::NNTP::Client::DEBUG ||= $opt_d;
  $Net::NNTP::Server::DEBUG ||= $opt_d;
  $News::NNTPAuth::DEBUG    ||= $opt_d;
}
if ($opt_p && $opt_p =~ /^(\d+)$/) { $PORT  = $1; } 	

# Create the server, and make it listen for a connection
my $SERVER = new Net::NNTP::Server()
	|| die "Couldn't start the server: $!\n";
$SERVER->openport($PORT) || die "Couldn't listen on port $PORT: $!\n";

# Generate the servers list 
foreach my $server (@CLIENTS) {
  # Load up the authentication information with NNTPAuth, if possible
  my ($nntpuser, $nntppass) = News::NNTPAuth->nntpauth($server->server);  
  $server->user($nntpuser) if $nntpuser;  
  $server->pass($nntppass) if $nntppass;

  # Connect to the news server now, or forever hold your peace.
  $server->connect ? $SERVER->push($server)
  		   : warn "Couldn't connect to " . $server->name . ": $!\n";
}


# Accept and process the connections
while (my $CLIENT = $SERVER->connect) {
  while ($CHILDREN >= $MAXKIDS) { sleep 1 }  # Only allow $MAXKIDS servers
  next unless defined fileno($CLIENT);		# Socket not open -> stop
  next unless my $ip = validate($CLIENT);
  
  # Fork off more clients; if we're above $MAXKIDS, though, wait for a
  # while and try again.  (This ought to have a real timeout, I guess)
  unless (my $count = _fork($CLIENT)) {
    next unless defined $count;
    $CHILDREN++; warn "Clients: $CHILDREN\n" if $DEBUG;
    next;
  }

  $SERVER->closeport; 	# child closes unused handle
  select($CLIENT);
  $| = 1;		# autoflush

  my $servers = 0;
  # Reconnect to the servers, if we're not already connected to them
  foreach my $server (@{$SERVER->newsservers}) {
    $server->reconnect && $servers++;
  }
  # If we couldn't connect, quit in disgust
  unless ($servers) {
    print $CLIENT "505 Couldn't connect to other servers\n";
    select(\*STDOUT);
    exit(0);
  }

  # Put the server into reader mode.
  $SERVER->process($CLIENT, 'mode reader');

  # Central loop - every time there's input, do something with it.
  while (<$CLIENT>) {		
    chomp;  
    next unless (defined($_) && $_ =~ /\S/);   
    print STDOUT "$ip: $_\n" if $DEBUG;
    last unless defined fileno($CLIENT);	# Socket not open -> stop
    $SERVER->process($CLIENT, $_) || last;
  }

  # Close up and exit.
  close $CLIENT;
  select(\*STDOUT);
  $SERVER->process($CLIENT, 'quit');
  exit 0;
}

$SERVER->closeport;
exit 0;

###############################################################################
### Functions #################################################################
###############################################################################

# Make sure that the connection is valid.  Code taken from Programming 
# Perl 3rd Edition (mostly).  Not yet done, really; I'm not sure what 
# kind of behaviour this is going to become, or where it's going to sit.
use Socket;
sub validate {
  my $socket = shift || return undef;
  return undef unless defined fileno($socket);	# Not an open socket -> skip it
  my $peername = getpeername($socket) || return undef;  
  my ($port, $iaddr) = unpack_sockaddr_in($peername);
  my $ip = inet_ntoa($iaddr);
  my $hostname = gethostbyaddr($iaddr, AF_INET);
  warn "Connection from $hostname ($ip)\n" if $DEBUG;
  # This should be a lot more sophisticated, and it will be
  unless (grep /^$hostname$/, @MACHINES) {
    print $socket "502 Permission Denied\n";
    close $socket;
    return undef;
  }
  $ip;
}

### _fork ( FILEHANDLE )
# Just the fork code.  
sub _fork {
  my $fh = shift;  return undef unless $fh && defined fileno($fh);
  if (my $pid = fork) {         # Success, and we're the parent
    close $fh;                  #   -> close the fh, return 0, and
    return 0;                   #   wait for the next call
  } elsif ( defined $pid ) {    # Success, and we're a child
    return 1;                   #   -> return something, so we can keep going
  } elsif ( $! == EAGAIN ) {    # Retry - this is the "do again" thing
    sleep 5;
    return _fork($fh);
  } else {
    die "Cannot fork: $!\n";
    return undef;
  }
}

### reaper ( PID ) 
# Takes care of cleaning up the children processes.  Probably doesn't work
# perfectly; it's not overly well tested or anything.  Please report bugs
# with this, and fixes for them, ASAP.  Thanks, kind reader.  
sub reaper { 
  my $kid = waitpid(-1, WNOHANG);
  if ($kid == -1) { 		# no child waiting; ignore
  } elsif (WIFEXITED($?)) {	# Process exited
    $CHILDREN--;
    warn "Client reaped, $CHILDREN remain\n" if $DEBUG;
    reaper();			# Make sure there's no more left
  } else {  			# False alarm; ignore
  }
}

### Usage() 
# Prints usage information and exits.

sub Usage {
  warn <<EOM;

$0 v$VERSION
an NNTP reader and server
Usage: $0 [-hvd] [-p port]

A news server written in perl, which gets its information from other news
servers.  Its primary purpose is to allow a standard newsreader to read
news from more than one site.  More information is available in its manual
page.

	-h		Print this message and exit.
	-v 		Print the version number and exit.
	-d		Print debugging information.
	-p port		Bind to 'port' instead of the default ($PORT)
			  (Note that the standard news port is 119; 
			   you must be root to bind there, though.)
EOM

  exit(0);
}

### Version()
# Prints version information and exits.

sub Version {
  warn "$0 v$VERSION\n";
  exit(0);
}

=head1 REQUIREMENTS

Requires Perl 5 or better, Net::NNTP, News::Article (for Net::NNTP::Server), 
Errno, News::NNTPAuth, Net::NNTP::Server, and Net::NNTP::Client.  Note that 
the last three come with the standard distribution.


=head1 SEE ALSO

L<Net::NNTP>, L<Net::NNTP::Server>, L<Net::NNTP::Client>

=head1 TODO

o Add in the rest of the functionality of INN (authinfo, help, ihave, list,
newgroups, newnews, slave, xgtitle, xhdr, xpat, xpath) into Net::NNTP::Server.  
(This realy belongs in with Net::NNTP::Server).

o Create a real admin interface, with authinfo and such, so that you can log
into a server and automatically get a list of groups rather than having to
set these variables here on your own.  

o Make this thing more of a daemon, so that you can start it and forget it.  

o Write some better code for handling the children.  

o Implement real authentication, and NNTPS.

o Work on other related projects, like an NNTP cache.

=head1 AUTHOR

Tim Skirvin <tskirvin@killfile.org>

=head1 COPYRIGHT

Copyright 2000 Tim Skirvin <tskirvin@killfile.org>

This code may be used and/or distributed under the same terms as Perl
itself.  Note that this is alpha code; use it at your own risk.

=cut

# Version History
# v0.5a 	Wed Nov  8 16:39:29 CST 2000
#   First version that's ready for release.  Still needs to be better
#   driven as a daemon, and real authentication.
